#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "bitstream.h"
#include "headers.h"
#include "vlc.h"
#include "string.h"

/* Lookup tables for VLC */
#include "vlc_codes/add_inc.h"
#include "vlc_codes/p_type.h"
#include "vlc_codes/b_type.h"
#include "vlc_codes/mv.h"
#include "vlc_codes/cbp.h"
#include "vlc_codes/lumsize.h"
#include "vlc_codes/chromsize.h"
#include "vlc_codes/dctcoeff.h"

/*************************/
#include "reconstruct.h"
/*************************/
//#define EX_BITSTREAM
//#define DEBUG_MACROBLOCK
//#define DEBUG_BLOCK

macroblock *mb;
motion_vec_state *mv_st;
extern frame_t scratch, past_ref;
FILE *mvfile;
FILE *blockfile;

int decode_vlc(const codeword *table, bitstream *b, int numbits) {
  codeword tmp;

  tmp = table[peek_bits(b,numbits)];
  if(tmp.code == -1) {
    fprintf(stderr, "Bad VLC lookup!\n");
//    while(1);
    fclose(mvfile);
    fflush(stderr);
    fflush(blockfile);
    fclose(blockfile);
    exit(-1);
  }
  read_bits(b, tmp.size);
  return(tmp.code);
}

/* Some tables are too big to look up all at once.  So do it in two
   stages.  If the first stage returns with a constant REFER_NEXT,
   go to the second */
int decode_vlc_2stage(const codeword *table1, const codeword *table2, 
		      bitstream *b, int bits1, int bits2) {
  int value;

  value = decode_vlc(table1, b, bits1);
  if(value == REFER_NEXT)
    value = decode_vlc(table2, b, bits2);
  return(value);
}

/* Do the VLC decoding for a motion vector */
int decode_mv(bitstream *b, int residue, int full_pel, int *prev) {
  static int mv_r, mv_code, rsize, f, complement_r, little, big;
  static int max, min, newvec, recon;

    mv_code = decode_vlc_2stage(mv_table, mv_table2, b, 7, 5);
    if(mv_code) {
      if(read_bits(b,1)) mv_code = -mv_code;
      if(residue > 1) mv_r = read_bits(b, residue-1);
      else mv_r = 0;
    }
    rsize = residue - 1;
    f = 1 << rsize;

    if((f==1) || (!mv_code)) complement_r = 0;
    else complement_r = f - 1 - mv_r;

    little = mv_code*f;
    if(little) {
      if(little > 0) {
	little = little - complement_r;
	big = little - (32 * f);
      } else {
	little = little + complement_r;
	big = little + (32 * f);
      }
    } else big = 0;

    assert(little != f * 16);

    max = 16*f - 1;
    min = -16 * f;
    
    newvec = *prev + little;
    
    if((newvec <= max) && (newvec >= min))
      recon = *prev + little;
    else
      recon = *prev + big;

    *prev = recon;
    if(full_pel) recon <<= 1;

    return(recon);
}

/* This is the complicated VLC for DCT coefficients. First, it's
   a big table.  Second, there are two different forms of fixed-
   length coding */
int runlevel(int *buffer, int prev_coeff, bitstream *b) {
  int symbol;
  int run, level;

  /* Do the run-level lookup */
  symbol = decode_vlc_2stage(dctcoeff_table, dctcoeff_table2, b, 10, 9);
  
  if (symbol==EOB) return(EOB);	/* End of block symbol */

  if (symbol==ESCAPE) {		/* Fixed length escape code */
    run = read_bits(b,6);	/* Run */
    level = (int)((signed char)peek_bits(b,8));
    if((level == -128) || (level == 0)) { /* Escape for 8 more bits */
      level = read_bits(b,16);
      if(level & 0x8000)
	level |= 0xFF00;
    } else {
      read_bits(b,8);
    }
  } else {			/* Level is in lower 8 bits */
    level = symbol & 0xFF;
    run = (symbol >> 8) & 0xFF;
    if(read_bits(b,1)) level = -level;
  }

  assert(prev_coeff+run <= 64);

  /* Actually do the writing into the block buffer */
  buffer[prev_coeff + run] = level;
  
  return(run);
}

/* Parses a block of DCT coefficients, and stores it in the appropriate
   block buffer of the macroblock structure */
void parse_block(macroblock *mb, int i, bitstream *b) {
  int dc_size;
  int coeff_num=0, run;
  int differential;
  int quant_scale;

  quant_scale = mb->quant;
  memset(mb->blocks[i], '\0', 64*sizeof(int));

  if(!((mb->cbp >> (5-i)) & 1)) return; /* We may have skipped this one */

  if(mb->type & MACROINTRA_MASK) { /* Intra macroblock */
    coeff_num++;
    if(i<4) {			/* Luminance block */
      dc_size = decode_vlc(lumsize_table, b, 7);
      if(dc_size) {		/* Set DC differential */
	differential = read_bits(b, dc_size);
	if(differential & (1 << (dc_size-1)))
	  *(mb->blocks[i]) = differential;
	else
	  *(mb->blocks[i]) = (-1 << dc_size) | (differential + 1);
      }
    } else {			/* Chrominance block */
      dc_size = decode_vlc(chromsize_table, b, 8);
      if(dc_size) {		/* Set DC differential */
	differential = read_bits(b, dc_size);
	if(differential & (1 << (dc_size-1)))
	  *(mb->blocks[i]) = differential;
	else
	  *(mb->blocks[i]) = (-1 << dc_size) | (differential + 1);
      }
    }
    while((run = runlevel(mb->blocks[i], coeff_num, b)) != EOB)
      coeff_num += run + 1;
  } else {			/* Non-intra */
    if(peek_bits(b,1)) {	/* Just check first coeff table */
      read_bits(b,1);
      *(mb->blocks[i]) = read_bits(b,1) ? -1 : 1;
      coeff_num++;
    }
    else {			/* Do regular lookup */
      run = runlevel(mb->blocks[i], 0, b);
      coeff_num += run + 1;
    }
    /* Now read all the DCT coeffs */

    while((run = runlevel(mb->blocks[i], coeff_num, b)) != EOB) {
      coeff_num += run + 1;
//      iquant_inter_single(mb->blocks[i][coeff_num - 1]);
    }
  }

  for(coeff_num=0; coeff_num<64; coeff_num++)
    fprintf(blockfile, "%i ", mb->blocks[i][coeff_num]);
  fprintf(blockfile, "\n");
#ifdef DEBUG_BLOCK
  printf("Block: ");
  for(coeff_num=0; coeff_num<64; coeff_num++)
    printf("%i ", mb->blocks[i][coeff_num]);
  printf("\n");
#endif
}


void parse_macroblock(bitstream *b, pic_header *pic, int quant) {
  static int address = -1;
  static int add_inc;
  static int pic_ref = 0;
  int code, i;
  
  if(pic_ref != pic->tem_ref) {	/* New picture */
    address = -1;
    pic_ref = pic->tem_ref;
  }
  mb->address=0;
  mb->type=0;
  /* Initialize quant scale to value set in last slice header */
  mb->quant=quant;
  mb->cbp=0;
  
  /* Look for any escapes or stuffing */
  add_inc=0;
  do {
    code = decode_vlc_2stage(add_inc_table, add_inc_table2, b, 8, 6); 
    if(code == MACROBLOCK_ESCAPE)
      add_inc += 33;
  } while(code < 0);
  add_inc += code;
  address += add_inc;
  mb->address = address;
  if(add_inc != 1) {
    if(pic->type != 3) skip_mblock(add_inc-1);  /* Temporary !!!!!! ******/
     if(pic->type != 3)
       mv_st->fmv.right = mv_st->fmv.down = mv_st->bmv.right = 
	 mv_st->bmv.down = 0;
  }
  
  switch(pic->type) {
  case 1:			/* I frame */
    if(peek_bits(b, 1) == 1) { /* The VLC has only two cases */
      read_bits(b, 1);
      mb->type = 16;
    } else {
      read_bits(b, 2);
      mb->type = 17;
    }
    break;
  case 2:			/* P frame */
    mb->type = decode_vlc(p_type_table, b, 6);
    break;
  case 3:			/* B frame */
    mb->type = decode_vlc(b_type_table, b, 6);
    break;
  case 4:			/* D frame */
    fprintf(stderr, "D frames not supported\n");
    exit(-1);
    break;
  default:
    fprintf(stderr, "Ack! Bogus picture type\n");
  }

  /* mb.type now contains 5 binary pieces of info.  Get them out by 
     bitmasking */
  if(mb->type & MACROQUANT_MASK)
    mb->quant = read_bits(b, 5);
  
  if(mb->type & MACROINTRA_MASK) /* Reset previous MV to zero if intra block */
    mv_st->fmv.right = mv_st->fmv.down = 
      mv_st->bmv.right = mv_st->bmv.down = 0;

  if(mb->type & MACROMVF_MASK)	{ /* Uses forward MV */
    mb->fmv.right = decode_mv(b, pic->fmv_code, pic->fmv_full,
			      &mv_st->fmv.right);
    mb->fmv.down = decode_mv(b, pic->fmv_code, pic->fmv_full,
			     &mv_st->fmv.down);
    fprintf(mvfile, "%i %i\n", mb->fmv.right, mb->fmv.down);
  } else {
    mb->fmv.right = mb->fmv.down = 0;
    if(pic->type != 3) mv_st->fmv.right = mv_st->fmv.down = 0;
  }

  if(mb->type & MACROMVB_MASK)	{ /* Uses backward MV */
    mb->bmv.right = decode_mv(b, pic->bmv_code, pic->bmv_full, 
			      &mv_st->bmv.right);
    mb->bmv.down = decode_mv(b, pic->bmv_code, pic->bmv_full,
			     &mv_st->bmv.down);
//    printf("%i %i\n", mb->bmv.right, mb->bmv.down);
  } else {
    mb->bmv.right = mb->bmv.down = 0;
//    mv_st->bmv.right = mv_st->bmv.down = 0;
  }    

  if(mb->type & MACROPATTERN_MASK) /* Has coded block pattern info */ {
    mb->cbp = decode_vlc(cbp_table, b, 9);
//    fprintf(stderr, "%x ", mb->cbp);
  }
  else
    /* If it doesn't we skip all the blocks */
    mb->cbp = 0;

  /* If the macroblock is intra coded, all the blocks must be coded */
  if(mb->type & MACROINTRA_MASK) mb->cbp = 0x3f;

#ifdef DEBUG_MACROBLOCK
   printf("Macroblock: %i Type: %i, %x Q: %i MV: %i %i %i %i CBP: %x\n",
	 address, pic->type, mb->type,  mb->quant, mb->fmv.right,
	  mb->fmv.down, mb->bmv.right, mb->bmv.down, mb->cbp);
#endif  

  for(i=0; i<6; i++) {
    parse_block(mb, i, b);
  }
  
  /******************************/ 
#ifndef EX_BITSTREAM
  if (pic->type == 1 || pic->type == 2){
    process_macroblock(mb, scratch, past_ref);
  }
#endif
    /******************************/
}

void init_mb() {
  int i;

  /* We'll allocate a structure once and overwrite it each time.  It's
     faster */
  mb = (macroblock *) malloc(sizeof(macroblock));
  for(i=0; i<6; i++)
    mb->blocks[i] = (int *)malloc(64*sizeof(int));

  mv_st = (motion_vec_state *) malloc(sizeof(motion_vec_state));
  mv_st->fmv.right = mv_st->fmv.down = mv_st->bmv.right = 
    mv_st->bmv.down = 0;
  
  mvfile = fopen("mv.txt", "w");
  blockfile = fopen("/tmp/block2.txt", "w");
}
