/****************************************************************
 * Peter Boettcher <pwb@andrew.cmu.edu>
 * 18-796 Homework #6
 * 3/30/99
 *
 * motion_est.c: Half-Pel Motion Estimation (with MMX!)
 *        Usage: motion_est <infile> <outfile>
 ****************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "mmx.h"
#include "headers.h"

#define SIZE_X 352
#define SIZE_Y 288
//#define SIZE_X 176
//#define SIZE_Y 144

#define FULL_PEL
#define USE_MMX

typedef struct {
  int x, y;
  int arr;
  float mag;
} node;

unsigned char *rframe[4], *cframe;
extern unsigned char *past_ref;

node *search_order[4096];

#ifdef USE_MMX
unsigned short int *result;
#endif

/* Function used in quicksort */
int compare_function(const void *a, const void *b) {
  node *n1, *n2;

  n1 = (node *) *(node **)a;
  n2 = (node *) *(node **)b;

  return(n1->mag > n2->mag);
}

/* Fills an array with all the possible shifts, both full and
   half-pel, which are to be searched over.  Then sorts this array on
   magnitude.  This allows the entire search range to be searched in
   increasing order of magnitude, hopefully making the fast-out in the
   MAD calculation more effective */
void init_search_order() {
  int y, x, base, i;

  for(y=-16; y<=15; y++)
    for(x=-16; x<=15; x++) {
#ifndef FULL_PEL
      base = (y+16) * 128 + (x+16)*4;
      for(i=0; i<4; i++) {
	search_order[base + i] = (node *)malloc(sizeof(node));
	search_order[base + i]->x = x;
	search_order[base + i]->y = y;
	search_order[base + i]->arr = i;
      }

      search_order[base]->mag = (float)(x*x + y*y);	
      search_order[base + 1]->mag = ((abs(x)+0.5)*(abs(x)+0.5) + y*y);
      search_order[base + 2]->mag = (x*x+ (abs(y)+0.5)*(abs(y)+0.5));
      search_order[base + 3]->mag = ((abs(x)+0.5)*(abs(x)+0.5) +
					  (abs(y)+0.5)*(abs(y)+0.5));
#else
      base = (y+16) * 32 + (x+16);
      search_order[base] = (node *)malloc(sizeof(node));
      search_order[base]->x = x;
      search_order[base]->y = y;
      search_order[base]->mag = (float)(x*x + y*y);	
#endif

    }
#ifndef FULL_PEL
  qsort(search_order, 4096, sizeof(node *), compare_function);
#else
  qsort(search_order, 1024, sizeof(node *), compare_function);
#endif
}

/* Creates half-pel shifted versions of the reference frame, for
   all three cases */
void calc_halfpels() {
  int x, y;

  rframe[0] = past_ref;
#ifndef FULL_PEL
  for(y=0; y<SIZE_Y; y++)
    for(x=0; x<SIZE_X; x++) {
      rframe[1][y*SIZE_X + x] =
	(rframe[0][y*SIZE_X + x] + rframe[0][y*SIZE_X + x + 1] + 1) / 2;
      rframe[2][y*SIZE_X + x] =
	(rframe[0][y*SIZE_X + x] + rframe[0][(y-1)*SIZE_X + x] + 1) / 2;
      rframe[3][y*SIZE_X + x] =
	(rframe[0][y*SIZE_X + x] + rframe[0][(y-1)*SIZE_X + x] + 
	 rframe[0][y*SIZE_X + x + 1] + rframe[0][(y-1)*SIZE_X + x + 1] + 2) /4;
    }
#endif
}

/* Calculate the total distortion between two buffers
    length 16.  There are two versions: one for portable
    C, one for hand-written MMX code.  The MMX code makes
    the program about 200% faster */
#ifndef USE_MMX
int mad16(unsigned char *c_row, unsigned char *r_row) {
  register int mad=0;
  unsigned char *endc=c_row+16;

  while(c_row < endc)
    mad += abs(*(c_row++) - *(r_row++));
  
  return(mad);
}
#else
int mad16(unsigned char *c_row, unsigned char *r_row) {
  /* These are all hand-written MMX assembly macros (written
     by me) in mmx.h */
  MMX_CLR;			/* Reset accumulation reg */
  MMX_SUB16(*c_row, *r_row, *(c_row+8), *(r_row+8));
  MMX_ADDALL(result[0]);		/* Now add up the eight bytes */
  return((int) *result);
}
#endif

/* Calculate the MAD between two blocks.  This function
     takes two pointers to the upper left corners of the
     blocks */
int calc_mad(unsigned char *cf, unsigned char *rf, int best_mad) {
  int mad = 0, i;

  for(i=0; i<16; i++) {
    mad += mad16(cf, rf);
    if(mad >= best_mad) return mad; /* Fast out */
    cf += SIZE_X;
    rf += SIZE_X;
  }
  return(mad);
}

int block_search(int cx, int cy, motion_vec *mv) {
  int i, rx, ry, interp_array;
  int l_edge, r_edge, t_edge, b_edge;
  int mad=10000000, best_x=0, best_y=0, newmad, arr=0;
  int checkbounds;
  node *next_search;

  /* Find out if we're on any borders: It's better to do this
      once outside the loop than each time inside */
  l_edge = cx == 0;
  r_edge = cx >= SIZE_X - 16;
  t_edge = cy == 0;
//  b_edge = cy >= SIZE_Y - 16;
  b_edge = (cy >= SIZE_Y/16 - 1);
  checkbounds = l_edge || r_edge || t_edge || b_edge;

  //  i = open("debug.yuv",O_WRONLY|O_CREAT|O_TRUNC, 0666);
  //  write(i, cframe, SIZE_X*SIZE_Y);
  //  write(i, rframe[0], SIZE_X*SIZE_Y);
  //  close(i);
  /* Main loop: spiralx and spiraly contain the coordinates for a
      spiral starting at the center.  Smaller magnitude MV's are
      more likely to occur, so check near the center first.  A
      spiral doesn't *quite* go in order of magnitude, but it's
      close enough */
#ifdef FULL_PEL
  interp_array = 0;
  for(i=0; i<1024; i++) {
#else
  for(i=0; i<4096; i++) {
#endif
    next_search = search_order[i];
    rx = next_search->x;
    ry = next_search->y;
#ifndef FULL_PEL
    interp_array = next_search->arr;
#endif
    if(checkbounds)
      /* Don't do all these compares unless necessary */
      if((l_edge && (rx < 0)) || (r_edge && (rx > 0))
	 || (t_edge && (ry < 0)) || (b_edge && (ry > 0)))
	continue;
    /* Find the MAD for this particular offset */
    newmad = calc_mad(&cframe[cy*SIZE_X + cx],
		      &rframe[interp_array][(cy+ry)*SIZE_X + (cx+rx)], mad);
    //    printf("%i %i: %i\n", rx, ry, newmad);
    if (newmad < mad) {
      /* This one's better than the previous best, so save it */
      mad = newmad;
      best_x = rx;
      best_y = ry;
#ifndef FULL_PEL
      arr = interp_array;
#endif
    }
    if(!mad) break;		/* If it's zero, we're done! */
  }
#ifdef FULL_PEL
  mv->right = 2*best_x;
  mv->down = 2*best_y;
#else
  mv->right = 2*best_x + (arr & 1);
  mv->down = 2*best_y + ((arr & 2) >> 1);
#endif

#ifdef USE_MMX
  MMX_EXIT;
#endif
  return mad;
}

void init_motion_est() {

  init_search_order();

  cframe = (unsigned char *) malloc(SIZE_X*SIZE_Y*sizeof(unsigned char) + 16);
//  rframe[0] = (unsigned char *) malloc(SIZE_X*SIZE_Y*sizeof(unsigned char)+16);
  rframe[1] = (unsigned char *) malloc(SIZE_X*SIZE_Y*sizeof(unsigned char)+16);
  rframe[2] = (unsigned char *) malloc(SIZE_X*SIZE_Y*sizeof(unsigned char)+16);
  rframe[3] = (unsigned char *) malloc(SIZE_X*SIZE_Y*sizeof(unsigned char)+16);

#ifdef USE_MMX
  /* If we're using MMX, the buffers need to be 64-bit aligned */
  cframe =(unsigned char *)((unsigned long)(cframe+8) & (unsigned long)(~7));
  rframe[0] =(unsigned char *)((unsigned long)(rframe[0]+8) & (unsigned long)(~7));
  rframe[1] =(unsigned char *)((unsigned long)(rframe[1]+8) & (unsigned long)(~7));
  rframe[2] =(unsigned char *)((unsigned long)(rframe[2]+8) & (unsigned long)(~7));
  rframe[3] =(unsigned char *)((unsigned long)(rframe[3]+8) & (unsigned long)(~7));
  result = (unsigned short int*) malloc(16);
  result =(unsigned short *)((unsigned long)(result+8) & (unsigned long)(~7));
#endif
}
